#include "ili_cache.h"
#include <algorithm>
#include <nlp_macros.h>
#include <nlp_config.h>
#include <nlp_filesystem.h>
#include <nlp_time.h>
#include <assert.h>
using namespace std;


Cache* CacheNode::p_Cache = NULL;


// ================================================================
CacheNode::CacheNode (void)
{
	// p_Cache = NULL;
	p_Objects = NULL;

	p_StateCompressed = NULL;
	i_CompressedSize = 0;
	i_UncompressedSize = 0;
}



//																
CacheNode::~CacheNode (void)
{
	// we should not delete all the nodes the transition
	// graph points to, since other nodes may point to	
	// them as well...									

	// assert ((Cache*)0x1 != p_Cache);
	p_Cache->b_Modified = true;
	p_Cache->ForgetNode (this);
	// p_Cache = (Cache*)0x1;

	hmp_ActionToState.clear ();

	if (NULL != p_Objects)
		((EnvObject*)p_Objects)->DestroyObjects (p_Objects);

	p_Objects = NULL;
	i_ObjectCount = 0;
}


//																
String CacheNode::GetState (void)
{
	return ZlibWrapper::Uncompress (p_StateCompressed, i_CompressedSize, i_UncompressedSize);
}



//																
String CacheNode::GetNormalizedState (void)
{
	return Cache::ConvertToNormalizedState (GetState ());
}



//																
void CacheNode::AddChild (String& _rAction, CacheNode* _pChild)
{
	String sAction (_rAction);
	sAction.Strip ();

	pair <ActionToState_hmp_t::iterator, bool> pairInsert;
	pairInsert = hmp_ActionToState.insert (make_pair (sAction, _pChild));
	assert (true == pairInsert.second);

	p_Cache->b_Modified = true;
}



//																
void CacheNode::DiscardBranch (String& _rAction)
{
	String sAction (_rAction);
	sAction.Strip ();

	hmp_ActionToState.erase (sAction);
}



//																
CacheNode* CacheNode::FindChild (String& _rAction)
{
	String sAction (_rAction);
	sAction.Strip ();

	ActionToState_hmp_t::iterator	ite;
	ite = hmp_ActionToState.find (sAction);
	if (hmp_ActionToState.end () == ite)
		return NULL;
	return ite->second;
}



//																
void CacheNode::SetState (String& _rState)
{
	i_UncompressedSize = _rState.length () + 1;
	ZlibWrapper::Compress (_rState, &p_StateCompressed, &i_CompressedSize);

	p_Cache->UnpackState (_rState, &p_Objects, &i_ObjectCount);
}



//																
CacheNode* CacheNode::GetNode (String& _sAction, String& _sState)
{
	if ('*' == _sAction.at (0))
		return p_Cache->GetNodeForState (_sState);

	else
	{
		// check if we have previously taken this action in	
		// this state and received the same response ...	
		CacheNode* pPrevious = FindChild (_sAction);
		if (NULL != pPrevious)
		{
			if (_sState == pPrevious->GetState ())
				return pPrevious;

			else
			{
				String x (_sAction);
				x.Strip ();
				ofstream file ("error.log", ios::out | ios::app);
				file << "[ERROR]  Same action (" << x 
					 << ") taken at different times from this state "
					 	"results in different outcomes." << '\n'
					 << "-- old ------------------------------------" << '\n'
					 << pPrevious->GetState () << '\n'
					 << "-------------------------------------------" << '\n'
					 << "-- new ------------------------------------" << '\n'
					 << _sState << '\n'
					 << "-------------------------------------------" << '\n';
				file << "\n\n";
				file.close ();
				return NULL;
			}
		}

		// if we've never taken this action before, just	
		// create a new node ...							
		else
		{
			CacheNode* pNode = p_Cache->CreateNode (_sState);
			//  pNode->s_NormalizedState = Cache::ConvertToNormalizedState (_sState);
			return pNode;
		}
	}
}




//																
void CacheNode::Save (ostream& _rStream, 
					  CacheNodeToIndex_map_t& _rmapNodeToIndex,
					  CacheNodeToLinks_map_t& _rmapNodeToLinks)
{
	// The cache is a graph, so we need to check that we're not	
	// saving the same node multiple times ...					
	if (_rmapNodeToIndex.end () != _rmapNodeToIndex.find (this))
		return;

	// The node-node links are saved to file in terms of node	
	// indices, so here we create & remember the index values	
	long lIndex = _rmapNodeToIndex.size ();
	_rmapNodeToIndex.insert (make_pair (this, lIndex));

	// save node environment state information to file.			
	_rStream << "<><><><><><><><><><>" << endl;
	_rStream << lIndex << endl;
	_rStream << GetState () << endl;

	// Return link information through the link map.			
	_rmapNodeToLinks.insert (make_pair (this, &hmp_ActionToState));

	// Save this node's children ...							
	ITERATE (ActionToState_hmp_t, hmp_ActionToState, ite)
	{
		CacheNode* pNode = ite->second;
		if (NULL != pNode)
			pNode->Save (_rStream, _rmapNodeToIndex, _rmapNodeToLinks);
	}
}



//																
void CacheNode::DebugPrintActions (void)
{
	ITERATE (ActionToState_hmp_t, hmp_ActionToState, ite)
		cout << "->" << ite->first << "<-" << endl;
}



// ================================================================
Cache::Cache (void)
{
	p_Root = NULL;
	pthread_rwlock_init (&rwl_CacheAccess, NULL);
	b_Modified = false;
	b_ClearningCache = false;
}


Cache::~Cache (void)
{
	Clear ();
	p_Root = (CacheNode*)0x1;
	pthread_rwlock_destroy (&rwl_CacheAccess);
}


//																
void Cache::ForgetNode (CacheNode* _pNode)
{
	if (true == b_ClearningCache)
		return;
	set_AllNodes.erase (_pNode);
}


//																
void Cache::Clear (void)
{
	b_ClearningCache = true;
	ITERATE (CacheNode_set_t, set_AllNodes, ite)
		delete *ite;
	set_AllNodes.clear ();
	hmp_NormalizedStateToNode.clear ();
	p_Root = NULL;
	b_ClearningCache = false;
}



//																
void Cache::UnpackState (String& _rState, void** _ppObjects, int* _pObjectCount)
{
	EnvObject::UnpackState (_rState, _ppObjects, _pObjectCount);
}



//																
String Cache::ConvertToNormalizedState (const String& _rState)
{
	// split the state into seperate objects ...
	String_dq_t	dqObjects;
	((String&)_rState).Split (dqObjects, '\x03');

	// cut out only the temporally consistent	
	// parts of each object ...					
	ITERATE (String_dq_t, dqObjects, ite)
	{
		if (false == (*ite).Has ("\x01"))
			continue;
		String_dq_t	dqValues;
		(*ite).Split (dqValues, '\x01', 3);

		String sObject = dqValues [3];
		sObject.Replace ("\n", " ");
		sObject = sObject.substr (sObject.find ("\x04"));
		*ite = sObject;
	}

	// sort the objects...						
	sort (dqObjects.begin (), dqObjects.end ());

	// join objects back together into a string	
	String sNormalizedResponse;
	ITERATE (String_dq_t, dqObjects, ite)
	{
		if ("" != sNormalizedResponse)
			sNormalizedResponse << "\x03""\n";
		sNormalizedResponse << *ite;
	}
	return sNormalizedResponse;
}



//																
CacheNode* Cache::CreateNode (String& _rState, bool _bSetModified)
{
	if (true == _bSetModified)
		b_Modified = true;

	CacheNode* pNode = new CacheNode;
	pNode->p_Cache = this;
	pNode->SetState (_rState);
	set_AllNodes.insert (pNode);
	return pNode;
}



//																
CacheNode* Cache::GetRoot (void)
{
	if (NULL == p_Root)
	{
		String sEmptyState ("[cache root]\x10");
		p_Root = CreateNode (sEmptyState);
	}
	return p_Root;
}



//																
CacheNode* Cache::FindNodeForState (const String& _sState)
{
	String sNormalizedState = Cache::ConvertToNormalizedState (_sState);
	NormalizedStateToNode_hmp_t::iterator	ite;
	ite = hmp_NormalizedStateToNode.find (sNormalizedState);
	if (hmp_NormalizedStateToNode.end () == ite)
		return NULL;

	return ite->second;
}



//																
CacheNode* Cache::GetNodeForState (String& _sState)
{
	String sNormalizedState = Cache::ConvertToNormalizedState (_sState);
	NormalizedStateToNode_hmp_t::iterator	ite;
	ite = hmp_NormalizedStateToNode.find (sNormalizedState);
	if (hmp_NormalizedStateToNode.end () != ite)
		return ite->second;

	CacheNode* pNode = CreateNode (_sState);
	hmp_NormalizedStateToNode.insert (make_pair (sNormalizedState, pNode));
	return pNode;
}



//																
void Cache::RememberNodeByNormalizedState (String& _sState, CacheNode* _pNode)
{
	hmp_NormalizedStateToNode.insert (make_pair (_sState, _pNode));
}



//																
void Cache::LockForReading (void)
{
	pthread_rwlock_rdlock (&rwl_CacheAccess);
}



//																
void Cache::LockForWriting (void)
{
	pthread_rwlock_wrlock (&rwl_CacheAccess);
}


//																
void Cache::Unlock (void)
{
	pthread_rwlock_unlock (&rwl_CacheAccess);
}



//																
bool Cache::Load (void)
{
	LockForWriting ();

	IndexToNode_hmp_t hmpIndexToNode;
	bool bRet = LoadThreadUnsafe (hmpIndexToNode);
	b_Modified = false;

	Unlock ();
	return bRet;
}



//																
bool Cache::Save (void)
{
	if (false == b_Modified)
		return true;

	String sTempPath = (config)"temp_cache_path";

	// remove old temporary files...			
	String_dq_t	dqOldFiles;
	if (true ==	Path::GetPathList (sTempPath, dqOldFiles))
	{
		if (false == dqOldFiles.empty ())
		{
			String_set_t setOldFiles;
			setOldFiles.insert (dqOldFiles.begin (), dqOldFiles.end ());

			dqOldFiles.clear ();
			ITERATE (String_set_t, setOldFiles, ite)
			{
				String sFileName (*ite);
				if (false == sFileName.Has ("/env.cache.tmp."))
					continue;
				dqOldFiles.push_back (*ite);
			}

			if (false == dqOldFiles.empty ())
				dqOldFiles.pop_back ();

			if (false == dqOldFiles.empty ())
			{
				ITERATE (String_dq_t, dqOldFiles, ite)
					Path::RemoveFile (*ite);

				cout << "Deleted " << dqOldFiles.size () 
					 << " old temporary files." << endl;
			}
		}
	}


	// lock cache & save to temporary file ...	
	LockForWriting ();
	cout << "Saving Cache; " << hmp_NormalizedStateToNode.size () 
		 << " nodes @ " << Time::sDateTime () << endl;

	String sTempCacheFileName;
	bool bRet = SaveInternal (sTempCacheFileName);
	Unlock ();
	
	// SaveInternal returns false only if		
	// the cache is uninitialized, so there's	
	// nothing to save...						
	if (false == bRet)
		return true;
	
	// move temporary file to actual name...	
	String sCacheFileName = (config)"cache_file_name";
	if (true == Path::Exists (sCacheFileName))
	{
		if (false == Path::RemoveFile (sCacheFileName))
			return false;
	}
	if (false == Path::MoveFile (sTempCacheFileName, sCacheFileName))
		return false;

	b_Modified = false;
	return true;
}



//																
bool Cache::LoadThreadUnsafe (IndexToNode_hmp_t& _hmpIndexToNode)
{
	File file;
	if (false == Path::Exists ((config)"cache_file_name"))
	{
		cout << "[INFO]  Cache file does not exist, will be created anew." << endl;
		return true;
	}
	if (false == file.Open ((config)"cache_file_name"))
		return false;

	String sLine;
	// IndexToNode_hmp_t hmpIndexToNode;

	int iDuplicates = 0;
	// read environment state information of nodes ...	
	String sID;
	String sState;
	String sPreviousLine;
	bool bErroneousTerminator = false;
	while (true == file.ReadLine (sLine))
	{
		bool bPreviousLineHad10 = sPreviousLine.Has ("\x10");
		sLine.Strip ();
		if ("" == sLine)
			continue;
		sPreviousLine = sLine;

		bool bEndAllStates = sLine.Has ("<<>><<>><<>>");
		bool bEndThisState = sLine.Has ("<><><><><><>");

		if ((false == bEndThisState) && (false == bEndAllStates))
			bErroneousTerminator |= bPreviousLineHad10;

		if ((true == bEndThisState) || (true == bEndAllStates))
		{
			bool bStateIsCorrupted = bErroneousTerminator;
			bErroneousTerminator = false;

			if ("" != sID)
			{
				sState.Strip ();
				if (("" != sState) && (true == sState.Has ("\n")) && (false == bPreviousLineHad10))
				{
					cout << "[..] End-of-state character x10 not found on previous line.  Ignoring state " << sID << endl;
					sID = "";
					sState = "";
					continue;
				}
				if (true == bStateIsCorrupted)
				{
					cout << "[..] State had terminator in the middle.  Ignoring state " << sID << endl;
					sID = "";
					sState = "";
					continue;
				}

				// We might have empty states saved to a	
				// learner's local cache in case it had		
				// previously operated in known-world-only	
				// mode.  These states must not be loaded	
				// up again, or they will stop normal		
				// operation by returning incorrect deadends
				if ("\x10" == sState)
				{
					cout << "[..] Ignoring empty state : " << sID << endl;
					sID = "";
					sState = "";
					continue;
				}
				if ("restart complete.\x10\n" == sState)
				{
					cout << "[..] Replacing 'restart complete.' with 'action completed, env has changed' : " << sID << endl;
					sState = "action completed, env has changed.\x10\n";
				}

				// We shouldn't remember error states!		
				// They shouldn't get into a saved cache	
				// in the first place, but in any case...	
				if ((true == sState.StartsWith ("[ERROR]")) ||
					(true == sState.StartsWith ("[[ERROR]]")))
				{
					cout << "[WARNING] Ignoring state : ->" << sState << "<-" << endl;
					sID = "";
					sState = "";
					continue;
				}

				String sNormalizedState = Cache::ConvertToNormalizedState (sState);

				CacheNode* pNode = NULL;
				if (true == sState.Has ("\x01"))
				{
					pNode = FindNodeForState (sState);
					if (NULL != pNode)
					{
						++ iDuplicates;
						cout << "   [**] " << sID << endl;
					}
				}

				if (NULL == pNode)
				{
					pNode = CreateNode (sState, false);
					RememberNodeByNormalizedState (sNormalizedState, pNode);
				}
				_hmpIndexToNode.insert (make_pair (sID, pNode));	
			}
			sID = "";
			sState = "";

			if (true == bEndAllStates)
				break;
			continue;
		}

		if ("" == sID)
		{
			sID = sLine;
			continue;
		}

		if ("" != sState)
			sState += "\n";
		sState += sLine;
	}

	if (0 != iDuplicates)
		cout << "      Cache : duplicates : " << iDuplicates << '/' 
			 << hmp_NormalizedStateToNode.size () << endl;

	// read node-node link information ...		
	String sFrom;
	while (true == file.ReadLine (sLine))
	{
		sLine.Strip ();
		if ("" == sLine)
			continue;

		if (true == sLine.Has ("from : "))
		{
			sFrom = sLine.substr (7);
			sFrom.Strip ();
			continue;
		}

		if (true == sLine.Has ("to : "))
		{
			sLine = sLine.substr (5);
			String_dq_t dqValues;
			sLine.Split (dqValues, '\x09');
			String sTo = dqValues [0];
			sTo.Strip ();
			String sAction = dqValues [1];
			sAction.Strip ();

			IndexToNode_hmp_t::iterator	iteFrom;
			iteFrom = _hmpIndexToNode.find (sFrom);
			if (_hmpIndexToNode.end () == iteFrom)
			{
				cout << "[ERROR]  Missing node : " << sFrom
					 << "  to : " << sTo
					 << "  action : " << sAction
					 << endl;
				continue;
			}

			IndexToNode_hmp_t::iterator	iteTo;
			iteTo = _hmpIndexToNode.find (sTo);
			if (_hmpIndexToNode.end () == iteTo)
			{
				cout << "[ERROR]  Missing node : " << sFrom
					 << "  to : " << sTo
					 << "  action : " << sAction
					 << endl;
				continue;
			}

			CacheNode* pFrom = iteFrom->second;
			CacheNode* pTo = iteTo->second;
			pFrom->hmp_ActionToState.insert (make_pair (sAction, pTo));
		}
	}

	// set root node ...						
	IndexToNode_hmp_t::iterator	iteRoot;
	iteRoot = _hmpIndexToNode.find ("0");
	assert (_hmpIndexToNode.end () != iteRoot);
	p_Root = iteRoot->second;

	cout << "      Cache : loaded " << _hmpIndexToNode.size () << " states." << endl;
	return true;
}



//																
bool Cache::SaveInternal (String& _rFileName)
{
	CacheNodeToIndex_map_t	mapNodeToIndex;
	return SaveThreadUnsafe (_rFileName, mapNodeToIndex);
}


//																
bool Cache::SaveThreadUnsafe (String& _rFileName, CacheNodeToIndex_map_t& _mapNodeToIndex)
{
	if (NULL == p_Root)
		return false;

	CacheNodeToLinks_map_t	mapNodeToLinks;


	time_t t = time (NULL);
	char zTime [201];
	#ifndef NDEBUG
	size_t iTimeLen = strftime (zTime, 200, "%Y-%m-%d-%H-%M-%S", localtime (&t));
	assert (iTimeLen < 200);
	#else
	strftime (zTime, 200, "%Y-%m-%d-%H-%M-%S", localtime (&t));
	#endif

	String sTempPath = (config)"temp_cache_path";
	_rFileName << sTempPath << "/env.cache.tmp." << zTime;

	File file (_rFileName, ios_base::out);
	// save environment state info for nodes ...	
	p_Root->Save (file, _mapNodeToIndex, mapNodeToLinks);

	file << "<<>><<>><<>><<>><<>><<>>" << endl;

	// save node-node link info ...					
	ITERATE (CacheNodeToLinks_map_t, mapNodeToLinks, iteLink)
	{
		CacheNode* pFrom = (CacheNode*) iteLink->first;
		ActionToState_hmp_t* phmpActionToState = iteLink->second;
		if (true == phmpActionToState->empty ())
			continue;

		CacheNodeToIndex_map_t::iterator	iteFrom;
		iteFrom = _mapNodeToIndex.find (pFrom);
		file << "from : " << iteFrom->second << endl;

		ITERATE (ActionToState_hmp_t, (*phmpActionToState), ite)
		{
			CacheNode* pTo = ite->second;
			String sAction = ite->first;
			sAction.Strip ();

			CacheNodeToIndex_map_t::iterator	iteTo;
			iteTo = _mapNodeToIndex.find (pTo);
			file << "to : " << iteTo->second << "\x09" << sAction << endl;
		}
	}

	file.flush ();
	file.close ();

	return true;
}



size_t Cache::GetNodeCount (void)
{
	return hmp_NormalizedStateToNode.size ();
}














