#ifndef __CACHE_LIB__
#define __CACHE_LIB__

#include <hash_map>
#include <deque>
#include <set>
#include <map>
#include <iostream>
#include <nlp_string.h>
#include <nlp_vector.h>
#include <zlib.h>
using namespace std;



// #define TRACE_PRINT_FUNCTIONS

#ifdef TRACE_PRINT_FUNCTIONS
	#define traceout	cout
#else
	#define traceout	while (false) cout
#endif


#define FEATURE_COUNT		20


class CompressedKey;
class EnvObject;
class Cache;
class CacheNode;
typedef Vector <char>								Feature_vec_t;
// typedef deque <EnvObject*>							EnvObject_dq_t;
typedef deque <CacheNode*>							CacheNode_dq_t;
typedef hash_map <String, CacheNode*>				ActionToState_hmp_t;
typedef hash_map <CompressedKey, CacheNode*>		NormalizedStateToNode_hmp_t;
typedef pair <pair <String, String>, String>		ActionState_pair_t;
typedef deque <ActionState_pair_t>					History_dq_t;
typedef pair <String, CacheNode*>					ActionCacheNode_pair_t;
typedef deque <ActionCacheNode_pair_t>				HistoryInCacheNodes_dq_t;
typedef hash_map <void*, long>						CacheNodeToIndex_map_t;
typedef hash_map <void*, ActionToState_hmp_t*>		CacheNodeToLinks_map_t;
typedef set <CacheNode*>							CacheNode_set_t;
typedef map <String, CacheNode*>					IndexToNode_hmp_t;



// -----------------------------------------------------
class CompressedKey
{
	private:
		char*	p_Data;
		int		i_CompressedLength;

	public:
		CompressedKey (void);
		CompressedKey (const CompressedKey& _rKey);
		CompressedKey (const String& _rString);
		~CompressedKey (void);

		bool operator< (const CompressedKey& _rRight) const;
		bool operator== (const CompressedKey& _rRight) const;

		static size_t Hash (const CompressedKey& _rKey)
		{
			unsigned long __h = 0; 
			for (int i = 0; i < _rKey.i_CompressedLength; ++ i)
				__h = 5*__h + _rKey.p_Data [i];

			return size_t(__h);
		};
};

namespace __gnu_cxx {
template<> struct hash<CompressedKey>
{
	  size_t operator()(const CompressedKey& __k) const 
	  	{ return CompressedKey::Hash (__k); };
}; 
}


// -----------------------------------------------------
class ZlibWrapper
{
	public:
		static void Compress (const String& _rState,
							  char** _ppBuffer,
							  int* _pCompressedLength);
		static String Uncompress (const char* _pCompressed, 
								  int _iCompressedLength,
								  int _iUncompressedLength);
};



// -----------------------------------------------------
class FeatureBitVector
{
	protected:
		char	c_Features [3];		// we have only 23 environment features

	public:
		FeatureBitVector (void);

		void Clear (void);
		void SetFeatures (char* _zData);
		void SetFeature (int _iIndex, int _iValue);
		void GetFeatures (Feature_vec_t& _rvecFeatures);
		const int operator[] (int _iIndex) const;

		bool operator== (const FeatureBitVector& _rOther);
		bool operator!= (const FeatureBitVector& _rOther)
			{ return (false == operator== (_rOther)); };
};

ostream& operator<< (ostream& _rStream, const FeatureBitVector& _rFeatures);
String& operator<< (String& _rString, const FeatureBitVector& _rFeatures);


// -----------------------------------------------------
class EnvObject
{
	protected:
		EnvObject (void);

	public:
		static size_t		i_ObjectMemorySize;
		char				z_ID [24];
		String				s_TemporallyUniqueId;
		FeatureBitVector	bit_Features;

		EnvObject (char* _zData);
		virtual ~EnvObject (void);

		void SetData (char* _zData);
		virtual void DestroyObjects (void* _pObjects);

		static void UnpackState (String& _rState, void** _ppObjects, int* _pObjectCount);
};





// -----------------------------------------------------
class CacheNode
{
	friend class Cache;
	friend class CacheInterface;

	protected:
		ActionToState_hmp_t		hmp_ActionToState;

	protected:
		static Cache*			p_Cache;
		char*					p_StateCompressed;
		int						i_UncompressedSize;
		int						i_CompressedSize;

	public:
		void*					p_Objects;
		int						i_ObjectCount;


	public:
		CacheNode (void);
		virtual ~CacheNode (void);

		virtual void SetState (String& _rState);
		String GetState (void);
		String GetNormalizedState (void);
		CacheNode* FindChild (String& _rAction);
		virtual void AddChild (String& _rAction, CacheNode* _pChild);
		virtual void DiscardBranch (String& _rAction);
		virtual CacheNode* GetNode (String& _sAction, String& _sState);

		int GetObjectCount (void)
			{ return i_ObjectCount; };
		EnvObject* GetObject (int _iIndex)
			{ return (EnvObject*)((char*)p_Objects + _iIndex * EnvObject::i_ObjectMemorySize); };
		static EnvObject* GetObject (void* _pObjects, int _iIndex)
			{ return (EnvObject*)((char*)_pObjects + _iIndex * EnvObject::i_ObjectMemorySize); };
		Cache* GetCache (void)
			{ return p_Cache; };

		void Save (ostream& _rStream, 
				   CacheNodeToIndex_map_t& _rmapNodeToIndex,
				   CacheNodeToLinks_map_t& _rmapNodeToLinks);

		void DebugPrintActions (void);
};


// -----------------------------------------------------
class Cache
{
	friend class CacheNode;
	friend class CacheInterface;

	protected:
		pthread_rwlock_t				rwl_CacheAccess;
		CacheNode*						p_Root;
		NormalizedStateToNode_hmp_t		hmp_NormalizedStateToNode;
		CacheNode_set_t					set_AllNodes;
		bool							b_Modified;
		bool							b_ClearningCache;

		static String ConvertToNormalizedState (const String& _rState);
		bool LoadThreadUnsafe (IndexToNode_hmp_t& _hmpIndexToNode);
		virtual bool SaveInternal (String& _rFileName);
		bool SaveThreadUnsafe (String& _rFileName, CacheNodeToIndex_map_t& _mapNodeToIndex);

	public:
		Cache (void);
		virtual ~Cache (void);

		void ForgetNode (CacheNode* _pNode);
		void Clear (void);

		// virtual EnvObject_dq_t UnpackState (String& _rState);
		virtual void UnpackState (String& _rState, void** _ppObjects, int* _pObjectCount);
		virtual CacheNode* CreateNode (String& _rState, bool _bSetModified = true);
		virtual CacheNode* GetRoot (void);
		virtual CacheNode* FindNodeForState (const String& _sState);
		virtual CacheNode* GetNodeForState (String& _sState);
		virtual void RememberNodeByNormalizedState (String& _sState, CacheNode* _pNode);
		virtual bool Load (void);
		bool Save (void);
		size_t GetNodeCount (void);

		void LockForReading (void);
		void LockForWriting (void);
		void Unlock (void);
};









#define CACHE_LOCK_NONE		0
#define CACHE_LOCK_READ		1
#define CACHE_LOCK_WRITE	2


enum HistoryState_e
{
	hs_ok,
	hs_retry,
	hs_cache_mismatch,
	hs_comm_error
};


typedef deque <HistoryState_e>	HistoryState_dq_t;



// -----------------------------------------------------
class EnvironmentInterface
{
	public:
		virtual ~EnvironmentInterface (void) {};

		virtual bool SendReceive (String& _rMessage, String* _pResponse) = 0;
		virtual bool StartEpisode (String& _rStartingState) = 0;
		virtual void EndEpisode (void) = 0;
		virtual String GetEnvironmentId (void) = 0;
};



// -----------------------------------------------------
class CacheInterface
{
	protected:
		char						c_CacheLockMode;
		bool						b_SendReceiveDirect;
		CacheNode*					p_CurrentNode;

	private:
		Cache*						p_Cache;
		EnvironmentInterface*		p_EnvironmentInterface;
		String						s_RequiredEnvStatingState;

		CacheNode*					p_StartingNode;

		String						s_PreviousActionSent;
		String						s_PreviousActionRequested;
		String						s_LastLiveState;
		CacheNode*					p_LastCacheNode;

		HistoryInCacheNodes_dq_t	dq_HistoryFromCache;
		History_dq_t				dq_HistoryToSave;
		HistoryState_dq_t			dq_StateOfHistoryToSave;

		bool						b_InEpisode;

		String MakeEnvironmentCurrent (void);
		String ConvertActionToLatest (String& _sAction, 
									  CacheNode* _pCacheNode);
		String UpdateActionFromCache (String& _sAction, 
									  String& _sLastLiveState,
									  CacheNode* _pCacheNode,
									  bool _bLogError);

		bool SwitchFromCacheToDirect (void);
		bool SaveHistoryToCache (void);
		CacheNode* SaveActionStateToNode (CacheNode* _pNode, 
										  String& _rAction, 
										  String& _rState,
										  int _iStepsIntoSave);

		CacheNode* SendReceiveThroughCache (String& _sAction);
		String SendReceiveDirect (String& _sActionToSend, 
								  String& _sActionReqested, 
								  bool _bRetryAttempt = false);

	public:
		static int					i_ForegroundFeatureIndex;
		static int					i_NewlyForegroundFeatureIndex;
		static bool					b_Verbose;
		bool						b_KnownWorldOnly;


		CacheInterface (void);
		~CacheInterface (void);

		void Init (Cache* _pCache, EnvironmentInterface* _pEnvInterface);
		bool StartEpisode (String& _sStartingState);
		bool EndEpisode (void);
		String SendReceive (String& _sAction);
		CacheNode* SendReceive (String& _sAction, CacheNode_dq_t* _pdqNewNodes);
};





#endif

